import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter_widgets/flutter_widgets.dart';

import 'models.dart';
import 'graph_controller.dart';

export 'models.dart';
export 'graph_controller.dart';

const double kBarsPadding = 4;

/// SimpleBarGraph is a simplified bar graph module suitable for fairly large data
/// representation.
/// Groups of data is formed and provided to the [controller]. The controller is
/// responsible for notifying listeners of a data bar selection.
///
/// [graphHeight] is used to set just the height of the bar (with its primary label)
/// in layout. That is, the actual height of the graph is = [graphHeight] - `text label height`.
///
class SimpleBarGraph extends StatefulWidget {
  /// Provides data for representation and updates listeners with data selection
  final GraphController controller;

  /// Height of the graph. Controls the height of a bar with its label
  final double graphHeight;

  /// The width of an individual bar
  final double barWidth;

  /// Should the group labels show?
  final bool showGroupLabels;

  SimpleBarGraph({
    @required this.controller,
    this.graphHeight = 250,
    this.barWidth = 60,
    this.showGroupLabels = true,
  }) : assert(controller != null, "Controller must not be null");

  @override
  _SimpleBarGraphState createState() => _SimpleBarGraphState();
}

class _SimpleBarGraphState extends State<SimpleBarGraph> {
  final ItemScrollController _graphScrollController = ItemScrollController();

  final ItemPositionsListener _itemPositionsListener =
      ItemPositionsListener.create();

  final List<DataItem> _chart = [];

  int startIndex = 0;

  int endIndex = 0;

  @override
  void initState() {
    super.initState();

    widget.controller.addListener(() {
      setState(() {});
    });

    // normalize the data
    widget.controller.groups.forEach((_group) {
      _chart.addAll(_group.data);
    });

    if (widget.showGroupLabels) {
      _listenToScrolls();
    }

    Timer(Duration(seconds: 1), () {
      _graphScrollController.scrollTo(
        index: widget.controller.groups.length - 1,
        alignment: -1,
        duration: Duration(milliseconds: 200),
        curve: Curves.easeInSine,
      );
    });
  }

  void _listenToScrolls() {
    _itemPositionsListener.itemPositions.addListener(() {
      /// Find out which bar is at the left and right edge and update the labels.
      /// this is for fancy labelling purposes.

      /// EdgeThreshold is the space a bar can occupy. Visually, `edgeThreshold`
      /// begins from `0` to `(barWidth / canvasWidth)` in terms of fraction of the screen.
      /// Therefore the trailing part of the screen will begin from `1 - edgeThreshold` to `1`.
      /// Note that it is in terms of fractions or percentage.
      double edgeThreshold = widget.barWidth / context.size.width;
      double endSpace = 1 - edgeThreshold;

      double oldStartDelta = 0;
      double oldEndDelta = 0;

      int oldStart = startIndex;
      int oldEnd = endIndex;

      _itemPositionsListener.itemPositions.value
          .forEach((ItemPosition _itemPosition) {
        //this part changes the start group label
        if (_itemPosition.itemTrailingEdge <= edgeThreshold) {
          // use delta to compare
          double delta = edgeThreshold - _itemPosition.itemTrailingEdge;
          if (delta > oldStartDelta) {
            startIndex = _itemPosition.index;
            oldStartDelta = delta;
          }
        }

        if (_itemPosition.itemLeadingEdge <= edgeThreshold) {
          double delta = edgeThreshold - _itemPosition.itemLeadingEdge;
          if (delta > oldStartDelta) {
            startIndex = _itemPosition.index;
            oldStartDelta = delta;
          }
        }

        // this part changes the end group label
        if (_itemPosition.itemLeadingEdge >= endSpace) {
          double delta = _itemPosition.itemLeadingEdge - endSpace;
          if (delta > oldEndDelta) {
            endIndex = _itemPosition.index;
            oldEndDelta = delta;
          }
        }

        if (_itemPosition.itemTrailingEdge >= endSpace) {
          double delta = _itemPosition.itemTrailingEdge - endSpace;
          if (delta > oldEndDelta) {
            endIndex = _itemPosition.index;
            oldEndDelta = delta;
          }
        }
      });

      if (startIndex != oldStart || endIndex != oldEnd) {
        setState(() {});
      }
    });
  }

  @override
  Widget build(BuildContext context) {
    return Column(
      children: <Widget>[
        Container(
          height: widget.graphHeight,
          child: Bars(
            height: widget.graphHeight,
            barWidth: widget.barWidth,
            graphController: widget.controller,
            controller: _graphScrollController,
            itemPositionsListener: _itemPositionsListener,
            chart: _chart,
          ),
        ),
        SizedBox(
          height: widget.showGroupLabels ? 8 : 0,
        ),
        widget.showGroupLabels
            ? GroupLabels(
                graphController: widget.controller,
                barSpace: widget.barWidth,
                leadingData: _chart[startIndex],
                trailingData: _chart[endIndex],
              )
            : SizedBox()
      ],
    );
  }
}

class Bars extends StatelessWidget {
  final double height;
  final double barWidth;

  final GraphController graphController;

  final ItemScrollController controller;

  final ItemPositionsListener itemPositionsListener;

  final List<DataItem> chart;

  Bars(
      {this.height,
      this.barWidth,
      this.graphController,
      this.controller,
      this.itemPositionsListener,
      this.chart});

  @override
  Widget build(BuildContext context) {
    return ScrollablePositionedList.builder(
      itemScrollController: controller,
      itemCount: chart.length,
      itemPositionsListener: itemPositionsListener,
      itemBuilder: (context, index) {
        final DataItem data = chart[index];

        return SizedBox(
          height: height,
          child: Column(
            children: <Widget>[
              Expanded(
                child: _Bar(
                  heightRatio: (data.value / graphController.maxValue.value),
                  width: barWidth,
                  selected: data == graphController.value.dataItem,
                  active: graphController.isActive(data),
                  onSelect: () {
                    graphController.select(data);
                  },
                ),
              ),
              SizedBox(
                height: 8,
              ),
              Text(
                data.label,
                style: TextStyle(
                  fontSize: 16,
                ),
              ),
            ],
          ),
        );
      },
      scrollDirection: Axis.horizontal,
    );
  }
}

class _Bar extends StatelessWidget {
  final double heightRatio;
  final double width;
  final bool selected;
  final bool active;
  final VoidCallback onSelect;

  _Bar({
    this.heightRatio,
    this.width,
    this.selected = false,
    this.active = false,
    this.onSelect,
  });

  Color _getColor(BuildContext context) {
    if (this.selected) {
      return Theme.of(context).accentColor;
    }

    if (this.active) {
      return Colors.grey[400];
    }

    return Colors.grey[300];
  }

  @override
  Widget build(BuildContext context) {
    return GestureDetector(
      onTap: this.onSelect,
      child: Container(
        width: this.width,
        padding: EdgeInsets.symmetric(
          horizontal: kBarsPadding,
        ),
        child: CustomPaint(
          painter: _BarPainter(
            drawRatio: this.heightRatio,
            color: _getColor(context),
          ),
        ),
      ),
    );
  }
}

class _BarPainter extends CustomPainter {
  final double drawRatio;
  final Color color;

  _BarPainter({this.drawRatio, this.color});

  @override
  void paint(Canvas canvas, Size size) {
    final paint = Paint();
    paint.color = this.color;

    double _drawHeight = drawRatio * size.height;

    canvas.drawRect(
        Rect.fromLTWH(
          0,
          size.height - _drawHeight,
          size.width,
          _drawHeight,
        ),
        paint);
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return true;
  }
}

class GroupLabels extends StatelessWidget {
  final double barSpace;
  final GraphController graphController;
  final DataItem leadingData;
  final DataItem trailingData;

  GroupLabels(
      {this.barSpace,
      this.graphController,
      this.leadingData,
      this.trailingData});

  Widget _getLeadingLabel(BuildContext context) {
    DataGroup group = graphController.getBelongingGroup(leadingData);
    bool active =
        graphController.value != null && graphController.value.group == group;

    return Container(
      padding: EdgeInsets.symmetric(horizontal: 16, vertical: 8),
      decoration: BoxDecoration(
        borderRadius: BorderRadius.only(
          topRight: Radius.circular(24),
          bottomRight: Radius.circular(24),
        ),
        color: active
            ? Theme.of(context).accentColor.withOpacity(0.1)
            : Colors.grey[100],
      ),
      child: Text(
        group.title,
        style: TextStyle(
          color: active ? Theme.of(context).accentColor : Colors.grey,
          fontSize: 16,
        ),
      ),
    );
  }

  Widget _getTrailingLabel(BuildContext context) {
    if (leadingData == trailingData) {
      return SizedBox();
    }
    DataGroup group = graphController.getBelongingGroup(trailingData);
    bool active =
        graphController.value != null && graphController.value.group == group;

    return Container(
      padding: EdgeInsets.symmetric(
        horizontal: 16,
        vertical: 8,
      ),
      decoration: BoxDecoration(
        borderRadius: BorderRadius.only(
          topLeft: Radius.circular(24),
          bottomLeft: Radius.circular(24),
        ),
        color: active
            ? Theme.of(context).accentColor.withOpacity(0.1)
            : Colors.grey[100],
      ),
      child: Text(
        group.title,
        style: TextStyle(
          color: active ? Theme.of(context).accentColor : Colors.grey,
          fontSize: 16,
        ),
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    return Row(
      children: <Widget>[
        _getLeadingLabel(context),
        Expanded(
          child: SizedBox(),
        ),
        _getTrailingLabel(context),
      ],
    );
  }
}
